/* ===========================================================================
                 PARSER FOR A REGULAR GRAMMAR
==============================================================================
The input grammar is given as a string, named gg, of the form:
   terminal    ::= 'a'..'z'
   nonterminal ::= 'A'..'Z'
   rsides      ::= terminal '.' | terminal nonterminal (right recursion)
   symprod     ::= nonterminal  rsides
   grammar     ::= symprod | symprod grammar
The empty productions, also called e-rules, are not allowed. The given grammar
can be nondeterministic.
When writing the string gg which encodes the productions of the input grammar,
the productions relative to the same non-terminal need not be grouped
together. For instance, the string "Qb.PbQPa.QaQ" stands for the productions:
1.Q->b; 2.P->bQ; 3.P->a; 4.Q->aQ.
The function 'length' gives the length of a string.
The string to be parsed is in the string 'stp' of characters from 'a'..'z'.
The rsides of the form: nonterminal terminal | terminal '.' can always
be transformed in rsides of the form: terminal '.' | terminal nonterminal,
that is, left-recursion can be avoided in favour of right-recursion.
(see: Esercise 3.7 in Hopcroft-Ullmann: Formal Languages and Their Relation to
Automata. Addison Wesley. 1969.)
------------------------------------------------------------------------------
The value 9999 means either a vacuous position of a production for '.' or
the non existent next production of a nonterminal (because all productions for
that nonterminal have been already considered).
------------------------------------------------------------------------------
When printing the ancestorslist, each item is a pair: `number.nonterminal'.
The number identifies the production for the nonterminal within the string
'gg' which represents the grammar. This is used for denoting the productions
already explored. For any given nonterminal the productions are assumed to be
ordered from left to right according to their occurrence in 'gg'.
There is a global variable 'traceon'. If 'traceon' is set to 'true' then we
can trace various steps through the execution of the program.
=========================================================================== */
#include <iostream.h>
#include <string>      // not <string.h>. File needed for using strings.
typedef int Values;
#include "list_module.cpp"
/*----------------------------------------------------------------------------
Now the type 'list' is the type of a list of integers. Now, the functions:
head, tail, cons, member, and printlist are available.
The global variables are: gg, stp, lgg, lstp1, traceon.
--------------------------------------------------------------------------- */
string gg, stp; // gg: given grammar, stp: string to parse.
int  lgg,lstp1;
/*----------------------------------------------------------------------------
lstp1 is the length of the string to parse -1.
-1 is due to the fact that indexing in c++ begins from 0:
       stp=abcab  length(stp)=5
index:     01234  lstp1=4
--------------------------------------------------------------------------- */
bool traceon;
//----------------------------------------------------------------------------
//                    printing a given production
void pprint(int i)
{cout << i/3+1 << "." << gg[i] << "->" << gg[i+1];
if (gg[i+2] != '.' ) {cout << gg[i+2];} cout << "; ";}
//----------------------------------------------------------------------------
//                    printing a given grammar
void gprint() {int i=0; while (i<lgg) {pprint(i); i=i+3;}; cout << "\n";}
//----------------------------------------------------------------------------
//                    printing the ancestorlist
void printnelist(list L)
{if (L!=NULL)
    {if (tail(L)==NULL)
     {cout << (head(L)/3)+1 << "." << gg[head(L)];}
     else {cout << (head(L)/3)+1 << "." << gg[head(L)] << ", ";
           printnelist(tail(L)); };};}

void print(list L)
{if (L==NULL) {cout << "[]";} else
{cout << "["; printnelist(L); cout << "]"; };}
//----------------------------------------------------------------------------
//                    tracing
void trace(string s, list al, int pg, int ps)
{if (traceon)
     {cout << "\n\nAncestorslist:      "; print(al); cout << ";  ";
     cout << "\nProduction at hand: ";
     if (pg==9999) {cout << "none;";} else {pprint(pg);};
     cout << "\nCharacter at hand:  " << stp[ps]; cout << "   => " << s;}}
//----------------------------------------------------------------------------
//                    next production
int findnextprod(int p)
{char s=gg[p];
do {p=p+3;} while (!((p>lgg) || (gg[p]==s)));
if (p<=lgg) {return p;} else {return 9999;};}
//----------------------------------------------------------------------------
//                    first production
int findfirstprod(int p)
{char s=gg[p]; int i=0;
while ((gg[i]!=s) && (i<=lgg)) {i=i+3;};
if (i<=lgg) {return i;} else {return 9999;};}
// ---------------------------------------------------------------------------
//                    parsing
bool parse(list al, int pg, int ps)
{list al1=NULL; int n=0,h=0,ps1=0;
if ((al==NULL) && (pg==9999))
   {trace("Fail.",al,pg,ps); return false;}
else if (pg==9999)
     {al1=tail(al); h=head(al); n=findnextprod(h); ps1=ps-1;
      trace("Alternative production from the father.",al,pg,ps);
      return parse(al1,n,ps1);}
else if ((gg[pg+1]!=stp[ps]) ||
        ((gg[pg+2]=='.') && (ps!=lstp1)) ||
        ((gg[pg+2]!='.') && (ps==lstp1)))
       {trace("Alternative production.",al,pg,ps);
        n=findnextprod(pg); return parse(al,n,ps);}
else if ((gg[pg+2]=='.') && (ps==lstp1))
     {trace("Success.\n",al,pg,ps); return true;}
else  {al1=cons(pg,al); n=findfirstprod(pg+2); ps1=ps+1;
       trace("Go down the string.",al,pg,ps);
       return parse(al1,n,ps1);};
}
//----------------------------------------------------------------------------
void main()
{traceon=true;
gg="Pa.PbQPbP";           // example 0
stp="ba";                 // true
lgg=gg.length(); lstp1=stp.length()-1;
cout << "\nThe given grammar is: "; gprint();
char axiom=gg[0];
cout << "The axiom is " << axiom << ".";
list al=NULL; int pg=0; int ps=0; bool ans=parse(al,pg,ps);
cout << "\nThe input string\n  " << stp << "\nis ";
if (!ans) {cout << "NOT ";}; cout << "generated by the given grammar.\n";
}
/* input: output:  traceon=true.
------------------------------------------------------------------------------
          The given grammar is: 1.P->a; 2.P->bQ; 3.P->bP;
          The axiom is P.

          Ancestorslist:      [];
          Production at hand: 1.P->a;
          Character at hand:  b   => Alternative production.

          Ancestorslist:      [];
          Production at hand: 2.P->bQ;
          Character at hand:  b   => Go down the string.

          Ancestorslist:      [2.P];
          Production at hand: none;
          Character at hand:  a   => Alternative production
                                     from the father.

          Ancestorslist:      [];
          Production at hand: 3.P->bP;
          Character at hand:  b   => Go down the string.

          Ancestorslist:      [3.P];
          Production at hand: 1.P->a;
          Character at hand:  a   => Success.

          The input string
            ba
          is generated by the given grammar.
--------------------------------------------------------------------------- */
/* Other examples already prepared:
gg="PbPPa.";                     // example 1
stp="aba";                       // false
stp="bba";                       // true

gg="Pa.PaQQb.QbP";               // example 2
stp="ab";                        // true
stp="ababa";                     // true
stp="aaba";                      // false

gg="Pa.Qb.PbQQaQ";               // example 3
stp="baaab";                     // true
stp="baab";                      // true
stp="bbaaba";                    // false
--------------------------------------------------------------------------- */
